<?php


namespace Verifone\Hosted\Helper\Includes;
use Verifone\Hosted\Helper\Library\SimpleCurl;
use Verifone\Hosted\Helper\Interfaces\ApiMapper;
use Verifone\Hosted\Helper\Settings\VerifoneHostedCartSettings;
use Verifone\Hosted\Model\Verifone;
use Magento\Framework\App\CacheInterface;
use Psr\Log\LoggerInterface;
use Exception;

/**
 * Class Verifone_Token
 */
class VerifoneToken {

	private $username;
	private $password;
	private $client_id;
	private $redirect_uri;
	private $code_challenge;
	private $code_verifier;
	private $base_url;
	private $token;

	/**
	 * @var VerifoneHostedCartSettings
	 */
	private $settings;

	/**
	 * @var SimpleCurl
	 */
	private $simple_curl;

	/**
	 * @var LoggerInterface
	 */
	private $logger;

	/**
	 * @var CacheInterface
	 */
	private $cache;

	/**
	 * @var
	 */
	private $paymentMethod;

	/**
	 * The token is stored along with the expiration time in the meta data
	 * To get it back and to check the exp time we use a "divider" and explode the string
	 */
	const VERIFONE_TOKEN_SPLITTER = '___EXPIRE___';

	/**
	 * VerifoneToken constructor.
	 *
	 * @param VerifoneHostedCartSettings $settings
	 * @param ApiMapper                  $api_mapper
	 * @param SimpleCurl                 $simple_curl
	 * @param LoggerInterface            $logger
	 * @param CacheInterface             $cache
	 * @param Verifone                   $paymentMethod
	 */
	public function __construct(
		VerifoneHostedCartSettings $settings, ApiMapper $api_mapper, SimpleCurl $simple_curl, LoggerInterface $logger, CacheInterface $cache, Verifone $paymentMethod
	) {
		$this->cache = $cache;
		$this->username  = $settings->getApiUsername();
		$this->password  = $settings->getApiPassword();
		$this->client_id = $settings->getApiClientId();
		$this->settings = $settings;
		$this->simple_curl = $simple_curl;
		$this->paymentMethod = $paymentMethod;
		// yeah this seems to only work with localhost for whatever reason
		// either way its not relevant since we're only looking for the code
		// in the header response
		// really strange way to generate tokens if you ask me
		$this->redirect_uri = $api_mapper->getTokenRedirectUri();
		$this->base_url = $api_mapper->getToken();
		$this->logger = $logger;
		$this->code_verifier  = $this->generateCodeVerifier();
		$this->code_challenge = $this->calculateCodeChallenge( $this->code_verifier );
	}

	/**
	 * @return string
	 * @throws Exception
	 */
	private function getAuthEndpointToken() {
		$sso_token = $this->simple_curl
			->setOpt(CURLOPT_URL, $this->getAuthEndpoint())
			->setOpt(CURLOPT_POST, true)
			->setOpt(CURLOPT_RETURNTRANSFER, true)
			->setOpt(CURLOPT_SSL_VERIFYHOST, false)
			->setOpt(CURLOPT_SSL_VERIFYPEER, false)
			->setOpt(CURLOPT_HTTPHEADER, [
				'Content-Type: application/json',
				'X-OpenAM-Username: ' . $this->username,
				'X-OpenAM-Password: ' . $this->password,
				'Accept-API-Version:' . VERIFONE_ENV_HEADER_VERSION,
			]);

		if($this->settings->getCurlVerifySsl()) {
			$sso_token
				->setOpt(CURLOPT_SSL_VERIFYHOST, '2')
				->setOpt(CURLOPT_SSL_VERIFYPEER, true);
		}

		$sso_token = $sso_token->getResult();

		$sso_token = json_decode( $sso_token, true );
		if ( json_last_error() !== JSON_ERROR_NONE ) {
			$this->logger->error('Error message', ['exception' => sprintf( 'There was an error decoding the json response last known error was "%s", in file %s, at line %s. API response %s',
				json_last_error(), __FILE__, __LINE__, $sso_token )]);
			throw new Exception( 'There was an error in the application' );
		}

		if ( ! isset( $sso_token['tokenId'] ) ) {
			$this->logger->error('Error message', ['exception' => sprintf( 'There was an error in the json response from "%s". Token is not present in response, in file %s, at line %s', $this->getAuthEndpoint(), __FILE__, __LINE__ )]);
			throw new Exception( 'There was an error in the application' );
		}

		return trim( $sso_token['tokenId'] );
	}

	/**
	 * @return false|string
	 * @throws Exception
	 */
	private function getAuthzEndpointResponse() {
		$token       = $this->getAuthEndpointToken();
		$post_fields = sprintf( "response_type=code&client_id=%s&csrf=%s&redirect_uri=%s&decision=allow&code_challenge=%s&code_challenge_method=%s", $this->client_id, $token, $this->redirect_uri, $this->code_challenge, VERIFONE_ENV_CODE_CHALLENGE_METHOD );

		$auth_code_request = $this->simple_curl
			->setOpt(CURLOPT_URL, $this->getAuthzEndpoint())
			->setOpt(CURLOPT_POST, true)
			->setOpt(CURLOPT_RETURNTRANSFER, true)
			->setOpt(CURLOPT_HEADER, true)
			->setOpt(CURLOPT_POSTFIELDS, $post_fields)
			->setOpt(CURLOPT_SSL_VERIFYHOST, false)
			->setOpt(CURLOPT_SSL_VERIFYPEER, false)
			->setOpt(CURLOPT_HTTPHEADER, [ 'Content-Type: application/x-www-form-urlencoded' ])
			->setOpt(CURLOPT_COOKIE, VERIFONE_ENV_COOKIE_NAME . '=' . $token)
		;

		if($this->settings->getCurlVerifySsl()) {
			$auth_code_request
				->setOpt(CURLOPT_SSL_VERIFYHOST, '2')
				->setOpt(CURLOPT_SSL_VERIFYPEER, true);
		}

		$auth_code_request = $auth_code_request->getResult();

		$code_pos  = strpos( $auth_code_request, '?code=' );
		$iss_pos   = strpos( $auth_code_request, '&iss=' );
		$auth_code = substr( $auth_code_request, $code_pos + 6, ( $iss_pos - $code_pos - 6 ) );

		if ( false === $code_pos || false === $iss_pos || false === $auth_code ) {
			$this->logger->error('Error message', ['exception' => sprintf( 'There was an error in the response from "%s". "iss" is not present in response, in file %s, at line %s', $this->getAuthzEndpoint(), __FILE__, __LINE__ )]);
			throw new Exception( 'There was an error in the application' );
		}

		return $auth_code;
	}

	/**
	 * Create a new token or get an existing one
	 * @param $paymentMethod
	 *
	 * @return mixed|string
	 * @throws Exception
	 */
	public function getToken() {
		$this->token = $this->createToken();
		return trim( $this->token );
	}

	/**
	 * @param $property
	 *
	 * @return mixed
	 * @throws Exception
	 */
	private function getTokenProperty($property) {
		$decoded_token = $this->decode();
		if(!isset($decoded_token[$property])) {
			$this->logger->error('Error message', ['exception' => sprintf( 'There was an error decoding the token last known error was "%s", in file %s, at line %s. Current token: %s',
				json_last_error(), __FILE__, __LINE__, $this->token )]);
			throw new Exception( 'There was an error in the application' );
		}

		return $decoded_token[$property];
	}

	/**
	 * @return mixed
	 */
	private function decode() {
		if ( null === $this->token ) {
			throw new Exception( 'Cannot decode a non-existing token. Generate the token first, then decode it.' );
		}

		return json_decode( base64_decode(
			str_replace( '_', '/',
				str_replace( '-', '+',
					explode( '.', $this->token )[1] )
			)
		),
			true );
	}

	/**
	 * This is a cascading series of calls:
	 *  $this->getAuthEndpointToken()
	 *  $this->getAuthzEndpointResponse()
	 *  $this->createToken()
	 *
	 * @return mixed
	 * @throws Exception
	 */
	private function createToken() {
		$auth_code = $this->getAuthzEndpointResponse();

		$post_fields = sprintf( "grant_type=authorization_code&code=%s&client_id=%s&redirect_uri=%s&code_verifier=%s", $auth_code, $this->client_id, $this->redirect_uri, $this->code_verifier );
		$access_token = $this->simple_curl
			->setOpt(CURLOPT_URL, $this->getTokenEndpoint())
			->setOpt(CURLOPT_RETURNTRANSFER, true)
			->setOpt(CURLOPT_POST, true)
			->setOpt(CURLOPT_POSTFIELDS, $post_fields)
			->setOpt(CURLOPT_SSL_VERIFYHOST, false)
			->setOpt(CURLOPT_SSL_VERIFYPEER, false)
			->setOpt(CURLOPT_HTTPHEADER, [ 'Content-Type: application/x-www-form-urlencoded' ])
		;

		if($this->settings->getCurlVerifySsl()) {
			$access_token
				->setOpt(CURLOPT_SSL_VERIFYHOST, '2')
				->setOpt(CURLOPT_SSL_VERIFYPEER, true);
		}

		$access_token = $access_token->getResult();

		$access_token = json_decode( $access_token, true );
		if ( json_last_error() !== JSON_ERROR_NONE ) {
			$this->logger->error('Error message', ['exception' => sprintf( 'There was an error decoding the json response last known error was "%s", in file %s, at line %s. API response %s',
				json_last_error(), __FILE__, __LINE__, $access_token )]);
			throw new Exception( 'There was an error in the application' );
		}

		if ( ! isset( $access_token['access_token'] ) ) {
			$this->logger->error('Error message', ['exception' => sprintf( 'There was an error in the json response from "%s". Access Token is not present in response, in file %s, at line %s. API response %s', $this->getTokenEndpoint(), __FILE__, __LINE__, $access_token )]);
			throw new Exception( 'There was an error in the application' );
		}

		return trim( $access_token['access_token'] );
	}


	/**
	 * @return string
	 */
	private function getTokenEndpoint() {
		return sprintf( '%s/oauth2/realms/root/realms/%s/access_token', $this->base_url, VERIFONE_ENV_REALM );
	}

	/**
	 * @return string
	 */
	private function getAuthEndpoint() {
		return sprintf( '%s/json/realms/root/realms/%s/authenticate', $this->base_url, VERIFONE_ENV_REALM );
	}

	/**
	 * @return string
	 */
	private function getAuthzEndpoint() {
		return sprintf( '%s/oauth2/realms/root/realms/%s/authorize', $this->base_url, VERIFONE_ENV_REALM );
	}

	/**
	 * @return false|string
	 */
	private function generateCodeVerifier() {
		// code_verifier=$(openssl rand -base64 1200 | tr -dc 'a-zA-Z0-9' | fold -w 50 | head -n 1)
		$t = base64_encode( openssl_random_pseudo_bytes( 1200 ) );
		$q = '';
		for ( $i = 0; $i < strlen( $t ); $i ++ ) {
			if ( ( ord( $t[ $i ] ) > ord( 'a' )
			       && ord( $t[ $i ] ) < ord( 'z' ) )
			     || ( ord( $t[ $i ] ) > ord( 'A' )
			          && ord( $t[ $i ] ) < ord( 'Z' ) )
			     || ( ord( $t[ $i ] ) > ord( '0' )
			          && ord( $t[ $i ] ) < ord( '9' ) )

			) {
				$q .= $t[ $i ];
			}
		}

		return substr( $q, 0, 50 );
	}

	/**
	 * @param $code_verifier
	 *
	 * @return string|string[]
	 */
	private function calculateCodeChallenge( $code_verifier ) {
		//Generate PKCE Challenge from Verifier and convert / + = characters"
		//code_challenge=`echo -n ${code_verifier} | shasum -a 256 | cut -d " " -f 1 | xxd -r -p | base64 | tr / _ | tr + - | tr -d =`

		$code_challenge = base64_encode( hash( 'sha256', $code_verifier, true ) );

		$code_challenge = str_replace( '+', '-', $code_challenge );
		$code_challenge = str_replace( '/', '_', $code_challenge );
		$code_challenge = str_replace( '=', '', $code_challenge );

		return $code_challenge;
	}

}
